
import threading

from . import _impl
from .entry import NetworkTableEntry
from .messages import *

__all__ = ["BadMessageError", "StreamEOF", "NetworkTableConnection",
           "ReadManager", "PROTOCOL_REVISION"]

class StreamEOF(IOError):
    pass

class ReadStream:
    def __init__(self, f):
        self.f = f

    def read(self, size=-1):
        data = self.f.read(size)
        if size is not None and size > 0 and len(data) != size:
            raise StreamEOF("end of file")
        return data

    def readStruct(self, s):
        data = self.f.read(s.size)
        if len(data) != s.size:
            raise StreamEOF("end of file")
        return s.unpack(data)

class NetworkTableConnection:
    """An abstraction for the NetworkTable protocol
    """
    def __init__(self, stream, typeManager):
        self.stream = stream
        self.rstream = ReadStream(stream.getInputStream())
        self.wstream = stream.getOutputStream()
        self.typeManager = typeManager
        self.write_lock = _impl.create_rlock('write_lock')
        self.isValid = True

    def close(self):
        if self.isValid:
            self.isValid = False
            self.stream.close()

    def flush(self):
        with self.write_lock:
            self.wstream.flush()

    def sendKeepAlive(self):
        with self.write_lock:
            self.wstream.write(KEEP_ALIVE.getBytes())
            self.wstream.flush()

    def sendClientHello(self):
        with self.write_lock:
            self.wstream.write(CLIENT_HELLO.getBytes(PROTOCOL_REVISION))
            self.wstream.flush()

    def sendServerHelloComplete(self):
        with self.write_lock:
            self.wstream.write(SERVER_HELLO_COMPLETE.getBytes())
            self.wstream.flush()

    def sendProtocolVersionUnsupported(self):
        with self.write_lock:
            self.wstream.write(PROTOCOL_UNSUPPORTED.getBytes(PROTOCOL_REVISION))
            self.wstream.flush()

    def sendEntry(self, entryBytes):
        # use entry.getAssignBytes or entry.getUpdateBytes
        with self.write_lock:
            self.wstream.write(entryBytes)
    
    def read(self, adapter):
        messageType = self.rstream.read(1)
        if messageType == KEEP_ALIVE.HEADER:
            adapter.keepAlive()
            
        elif messageType == CLIENT_HELLO.HEADER:
            protocolRevision = CLIENT_HELLO.read(self.rstream)[0]
            adapter.clientHello(protocolRevision)
            
        elif messageType == SERVER_HELLO_COMPLETE.HEADER:
            adapter.serverHelloComplete()
            
        elif messageType == PROTOCOL_UNSUPPORTED.HEADER:
            protocolRevision = PROTOCOL_UNSUPPORTED.read(self.rstream)[0]
            adapter.protocolVersionUnsupported(protocolRevision)
            
        elif messageType == ENTRY_ASSIGNMENT.HEADER:
            entryName, (typeId, entryId, entrySequenceNumber) = \
                    ENTRY_ASSIGNMENT.read(self.rstream)
            entryType = self.typeManager.getType(typeId)
            if entryType is None:
                raise BadMessageError("Unknown data type: 0x%x" % typeId)
            value = entryType.readValue(self.rstream)
            adapter.offerIncomingAssignment(NetworkTableEntry(entryName, entryType, value, id=entryId, sequenceNumber=entrySequenceNumber))
            
        elif messageType == FIELD_UPDATE.HEADER:
            entryId, entrySequenceNumber = FIELD_UPDATE.read(self.rstream)
            entry = adapter.getEntry(entryId)
            if entry is None:
                raise BadMessageError("Received update for unknown entry id: %d " % entryId)
            value = entry.getType().readValue(self.rstream)
            adapter.offerIncomingUpdate(entry, entrySequenceNumber, value)
            
        else:
            raise BadMessageError("Unknown Network Table Message Type: %s" % (messageType))

class ReadManager:
    """A periodic thread that repeatedly reads from a connection
    """
    def __init__(self, adapter, connection, name=None):
        """create a new monitor thread
        
        :param adapter:
        :type  adapter: :class:`.ServerConncetionAdapter` or :class:`.ClientConnectionAdapter`
        :param connection:
        :type  connection: :class:`NetworkTableConnection`
        """
        self.adapter = adapter
        self.connection = connection
        self.running = True
        
        self.thread = threading.Thread(target=self.run, name=name)
        self.thread.daemon = True

    def start(self):
        self.thread.start()

    def stop(self):
        self.running = False
        try:
            self.thread.join()
        except RuntimeError:
            pass

    def run(self):
        while self.running:
            try:
                self.connection.read(self.adapter)
            except BadMessageError as e:
                self.adapter.badMessage(e)
            except IOError as e:
                self.adapter.ioError(e)




__all__ = ["create_rlock", 'sock_makefile']

import socket
import threading

def create_rlock(name):
    return threading.RLock()

def sock_makefile(s, mode):
    return s.makefile(mode)

def sock_create_connection(address):
    return socket.create_connection(address)


# Call this before creating any NetworkTable objects
def enable_lock_debugging():
    
    from . import _impl_debug
    
    g = globals()
    g['create_rlock'] = _impl_debug.create_tracked_rlock
    g['sock_makefile'] = _impl_debug.blocking_sock_makefile
    g['sock_create_connection'] = _impl_debug.blocking_sock_create_connection





'''
    Instrumentation for finding deadlocks in networktables    
'''

from __future__ import print_function

import inspect
import socket
import threading
import time


# List of locks that can be acquired from the main thread
main_locks = [
    'entry_lock',
    'trans_lock',
]

# List of locks that are allowed to be held when accessing a socket
# -> must never be locks that can be acquired by the main thread
sock_locks = [
    'client_conn_lock',
    'server_conn_lock',
    'write_lock',
]

# Dictionary of locks
# key: name, value: locks that can be held when acquiring the lock
locks = {
    # Never held by robot thread
    'client_conn_lock': [
        'client_conn_lock'
    ],
    
    'entry_lock': [
        'entry_lock',
        'client_conn_lock'
    ],
    
    # Never held by robot thread
    'server_conn_lock': [
        'server_conn_lock',       
    ],
    
    'trans_lock': [
        'entry_lock',
        
        # Not 100% sure if this should be allowed
        # -> this only happens when NetworkTable API calls are made from
        #    a fired listener
        'server_conn_lock',
        'client_conn_lock'
    ],
    
    # Never held by robot thread
    'write_lock': [
        'client_conn_lock',
        'server_conn_lock',
        'write_lock',
    ],
}

local = threading.local()

class WrappedLock(threading._PyRLock):
    
    def __init__(self, name):
        threading._PyRLock.__init__(self)
        self._name = name
        self._nt_creator = _get_caller()
    
    def acquire(self, blocking=True, timeout=-1):
        
        # This check isn't strictly true.. 
        if isinstance(threading.current_thread(), threading._MainThread):
            assert self._name in main_locks, "%s cannot be held in main thread" % self._name
        
        if not hasattr(local, 'held_locks'):
            local.held_locks = []
        
        for lock in local.held_locks:
            assert lock in locks[self._name], "Cannot hold %s when trying to acquire %s" % (lock._name, self._name)
        
        retval = threading._PyRLock.acquire(self, blocking=blocking, timeout=timeout)
        if retval != False:
            local.held_locks.append(self)

    __enter__ = acquire
    
    def release(self):
        threading._PyRLock.release(self)
        assert local.held_locks[-1] == self
        local.held_locks.pop()
    
    # Allow this to be used in comparisons
    
    def __eq__(self, other):
        if isinstance(other, str):
            return self._name.__eq__(other)
        else:
            return self._name.__eq__(other._name)
    
    def __cmp__(self, other):
        if isinstance(other, str):
            return self._name.__cmp__(other)
        else:
            return self._name.__cmp__(other._name)
    
    def __hash__(self):
        return self._name.__hash__()

def create_tracked_rlock(name):
    assert name in locks
    return WrappedLock(name)
    
def assert_not_locked(t):
    
    assert not isinstance(threading.current_thread(), threading._MainThread), \
        "Should not make socket calls from main thread"
    
    if not hasattr(local, 'held_locks'):
        local.held_locks = []
    
    for lock in local.held_locks:
        assert lock in sock_locks, \
            "ERROR: network %s was made while holding %s" % (t, lock._name)

class WrappedFile:
    def __init__(self, file):
        self._file = file
        
    def write(self, data):
        assert_not_locked('write')
        time.sleep(1)
        return self._file.write(data)
        
    def read(self, *args, **kwargs):
        assert_not_locked('read')
        time.sleep(1)
        return self._file.read(*args, **kwargs)
        
        
    def __getattr__(self, attr):
        return getattr(self._file, attr)


def blocking_sock_makefile(s, mode):
    return WrappedFile(s.makefile(mode))

def blocking_sock_create_connection(address):
    assert_not_locked('connect')
    time.sleep(1)
    return socket.create_connection(address)

def _get_caller():
    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 3)
    return '%s:%s %s' % (calframe[3][1], calframe[3][2], calframe[3][3])




#!/usr/bin/env python3
#
# This is a NetworkTables server (eg, the robot or simulator side).
#
# On a real robot, you probably would create an instance of the 
# wpilib.SmartDashboard object and use that instead -- but it's really
# just a passthru to the underlying NetworkTable object.
#
# When running, this will continue incrementing the value 'robotTime',
# and the value should be visible to networktables clients such as 
# SmartDashboard. To view using the SmartDashboard, you can launch it
# like so:
#
#     SmartDashboard.jar ip 127.0.0.1
#

import time
from networktables import NetworkTable

# To see messages from networktables, you must setup logging
import logging
logging.basicConfig(level=logging.DEBUG)

sd = NetworkTable.getTable("SmartDashboard")

i = 0
while True:
    try:
        print('dsTime:', sd.getNumber('dsTime'))
    except KeyError:
        print('dsTime: N/A')

    sd.putNumber('robotTime', i)
    time.sleep(1)
    i += 1





from networktables2 import StringArray
from .networktable import NetworkTable

class ChooserControl(object):
    '''
        Interacts with a :class:`wpilib.sendablechooser.SendableChooser`
        object over NetworkTables.
    '''
    
    def __init__(self, key, on_choices=None, on_selected=None):
        '''
            :param key: NetworkTables key
            :type  key: str
            :param on_choices: A function that will be called when the
                               choices change. Signature: fn(value)
            :param on_selection: A function that will be called when the
                                 selection changes. Signature: fn(value)
        '''
        
        self.subtable = NetworkTable.getTable('SmartDashboard').getSubTable(key)

        self.on_choices = on_choices
        self.on_selected = on_selected
        
        if on_choices or on_selected:
            self.subtable.addTableListener(self._on_change, True)

    def close(self):
        '''Stops listening for changes to the ``SendableChooser``'''
        if self.on_choices or self.on_selected:
            self.subtable.removeTableListener(self._on_change)
    
    def getChoices(self):
        '''
            Returns the current choices. If the chooser doesn't exist, this
            will return an empty array.
        
            :rtype: :class:`.StringArray`
        '''
        choices = StringArray()
        try:
            self.subtable.retrieveValue('options', choices)
        except KeyError:
            pass
        return choices
    
    def getSelected(self):
        '''
            Returns the current selection or None
        
            :rtype: str
        '''
        selected = self.subtable.getString('selected', None)
        if selected is None:
            selected = self.subtable.getString('default', None)
        return selected
    
    def setSelected(self, selection):
        '''
            Sets the active selection on the chooser
            
            :param selection: Active selection name
        '''
        self.subtable.putString('selected', selection)
    
    def _on_change(self, table, key, value, isNew):
        if key == 'options':
            if self.on_choices is not None:
                self.on_choices(value)
        elif key == 'selected':
            if self.on_selected is not None:
                self.on_selected(value)
        elif key == 'default':
            if self.on_selected is not None and self.subtable.getString('selected', None) is None:
                self.on_selected(value)
